/*
	Implementation of protocol T=0
	
	This file is part of Smartk, a smart card framework for the Linux Kernel.
	Copyright (C) 2003	Luigi Catuogno		<luicat@dia.unisa.it>
						Roberto Gassira'	<robgas@dia.unisa.it>
						Michele Masullo		<micmas@dia.unisa.it>
	
    Smartk is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Smartk is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "smartk.h"
#include "pps.h"
#include "apdu.h"				/* x APDU 		*/
#include <linux/mm.h>			/*x allocare memoria dinamica	rompe il cazzo con macro di open
									ma che merda di header del cazzo che hanno fatto sti froci!*/
extern void * kmalloc (size_t, int);

/* Definizione macro per licenza */
#define MY_AUTHOR       "Roberto Gassira' <robgas@tcfs.it>"
#define MY_DESC         "Modulo per implementazione protocollo T0"

/*	Macro per timeout carta	*/
#define PROTOCOL_T0_DEFAULT_WI         10
#define PROTOCOL_T0_MAX_SHORT_COMMAND  260
#define PROTOCOL_T0_MAX_SHORT_RESPONSE 258
#define PROTOCOL_T0_MAX_NULLS          200


/* Identificazione del codice e della sua licenza*/
MODULE_AUTHOR(MY_AUTHOR);
MODULE_DESCRIPTION(MY_DESC);
MODULE_LICENSE("GPL");

/*	Puntatore al core */
static struct smartk *mycore;
	
/*	Dichiarazione funzioni interne al modulo	*/
//BYTE * str_to_BYTE(char * data, int len);
static ProtocolParameters * ProtocolParameters_New (void);
static void ProtocolParameters_Perform(ProtocolParameters * pps, ATR * atr);
static int pt0_smart_set_timings();
static int pt0_smartk_check_begin_convention();
static int pt0_smartk_check_end_convention();
static int pt0_smartk_trasmit(BYTE *data, int len_data);
static int pt0_smartk_case1(APDU_Cmd *cmd, APDU_Rsp **rsp);
static int pt0_smartk_case2S(APDU_Cmd *cmd, APDU_Rsp **rsp);
static int pt0_smartk_case3S(APDU_Cmd *cmd, APDU_Rsp **rsp);	
static int pt0_smartk_case4S(APDU_Cmd *cmd, APDU_Rsp **rsp);
static int pt0_smartk_exchange_TPDU(APDU_Cmd *cmd, APDU_Rsp **rsp);	

/*	Funzioni Gestione modulo	*/
static int pt0_smartk_init(void);
static void pt0_smartk_exit(void);

/*	Funzioi Esportate	*/
static int pt0_smartk_init_card(struct smartk * core);
static int pt0_smartk_data(struct smartk * core, BYTE *new_cmd, int len_cmd, BYTE *new_rsp, int *len_rsp);
static int pt0_smartk_cleanup_card(struct smartk * core);
	
/*	Struttura per registrazione	*/
static struct pt_smartk t0 = {
	.activate_card = pt0_smartk_init_card,
	.data = pt0_smartk_data,
	.disactivate_card = pt0_smartk_cleanup_card
};

/*	Funzione per trasformare una stringa in un BYTE */
//~ BYTE * str_to_BYTE(char * data, int len){
	//~ BYTE * ret = (BYTE *)kmalloc(len * sizeof(BYTE), GFP_KERNEL);
	//~ int number = 0;
	//~ //PERROR("len = %d", len);
	//~ for(number = 0; number < len; number++){
		//~ ret[number] = (BYTE)data[number];
		//~ //PERROR("heheheh %02X\n", ret[number]);
	//~ }
	
	//~ return ret;
//~ }

/*	Funzioni prese da c.prados per PPS	*/
static ProtocolParameters * ProtocolParameters_New (void){
  ProtocolParameters *pps = (ProtocolParameters *)kmalloc(sizeof (ProtocolParameters), GFP_KERNEL);

  if (pps != NULL){
      pps->f = ATR_DEFAULT_F;
      pps->d = ATR_DEFAULT_D;
      pps->n = ATR_DEFAULT_N;
    }
 	return pps;
}

static void ProtocolParameters_Perform(ProtocolParameters * pps, ATR * atr){
  	ATR_GetParameter (atr, ATR_PARAMETER_N, &(pps->n));
    ATR_GetParameter (atr, ATR_PARAMETER_D, &(pps->d));
    ATR_GetParameter (atr, ATR_PARAMETER_F, &(pps->f));
    
	PDEBUG("ProtocolParameters_Perform: F=%d, D=%d, N=%d\n", \
			(int)pps->f, (int)pps->d, pps->n );	
}


/*	Calcolo dei valori di timings partendo dall'atr	*/
static int pt0_smart_set_timings(){
	BYTE WI;
	ATR * atr = mycore->atr;
	ProtocolParameters * pps;
	unsigned long WWT;
	/*	Calcolo dei valori PPS	*/
	pps = ProtocolParameters_New();
	ProtocolParameters_Perform(pps, atr);
	
	/*	Calcolo valore wi	*/
	if (ATR_GetInterfaceByte (atr, 2, ATR_INTERFACE_BYTE_TC, &(WI)) != ATR_OK){
    	WI = PROTOCOL_T0_DEFAULT_WI;
		PDEBUG("pt_t0:set_timings->Wi valore default = %d\n", WI);
	}
	else
		PDEBUG("pt_t0:set_timings->Wi valore calcolato = %d\n", WI);

	/*	Calcolo del Work Waiting Time	*/
	/* WWT = 960 * WI * (Fi / f) * 1000 milliseconds */
  	WWT = (long unsigned int) (960 * WI * (pps->f / mycore->ifd->clock_rate) * 1000);
	PDEBUG("pt_t0:set_timings->WWT= %ld\n", WWT);
	
	/*	Aggiornamento info card	*/
	mycore->timings.block_timeout = WWT;
	mycore->timings.char_timeout = WWT;
	mycore->timings.block_delay = mycore->timings.char_delay = 0;
}

/*	Controllo prima di inizio trasmissione dati */
static int pt0_smartk_check_begin_convention(){
	/*	Controllo su cambio convenzione	*/
	if(mycore->convention == ATR_CONVENTION_INVERSE)
		if( mycore->ifd->set_parity(mycore, IO_SERIAL_PARITY_ODD) != OK)
			return FAIL;			
	return OK;
}

/*	Controllo dopo aver trasmesso dati */
static int pt0_smartk_check_end_convention(){
	/*	Controllo su cambio convenzione	*/
	if(mycore->convention == ATR_CONVENTION_INVERSE)
		if( mycore->ifd->set_parity(mycore, IO_SERIAL_PARITY_EVEN) != OK)
			return FAIL;			
	return OK;
}

/*	Spedizione dei dati	con controllo convention	*/
static int pt0_smartk_trasmit(BYTE *data, int len_data){
	BYTE *buffer, *to_send;
  	int i;
	
	/*	Controllo inversione bit	*/
 	if (mycore->convention == ATR_CONVENTION_INVERSE){
		buffer = (BYTE *)kmalloc((sizeof (BYTE) * len_data), GFP_KERNEL);
      	memcpy (buffer, data, len_data);
		
		/*	Inversione dei dati	*/
		for(i=0; i < len_data; i++)
			buffer[i] =  ~(INVERT_BYTE (buffer[i]));

      	to_send = buffer;
    }
  	else
    	to_send = data;

	/*	Invio Dati	*/
	if( mycore->ifd->transmit(mycore, to_send, len_data) !=  OK ){
		PERROR("pt_t0:trasmit->Errore trasmissione dati\n");
		return FAIL;
	}
  	
	/*	Controllo pulizia memoria*/
 	if (mycore->convention == ATR_CONVENTION_INVERSE)
		kfree(buffer);

  return OK;
}

/*	Ricezione Dati	*/
static int pt0_smartk_receive(BYTE *data, int len_data){
	int i;
	PDEBUG("pt_t0:receive->Richiesta ricezione byte %d\n", len_data);
	if( mycore->ifd->receive(mycore, data, len_data) != OK )
		return FAIL;
	
	//PDEBUG("pt_t0:receive->Ricezione Stringa Prima conversione %s\n", mycore->xpdu_to_string(data, len_data));
	
	if( mycore->convention == ATR_CONVENTION_INVERSE ){
		/*	Inversione dei dati	*/
		for(i=0; i < len_data; i++)
			data[i] =  ~(INVERT_BYTE (data[i]));
	}
	
	/*	Output della chiamata	*/
	PDEBUG("pt_t0:receive->Ricezione Stringa dopo conversione %s\n", mycore->xpdu_to_string(data, len_data));
	return OK;
}


/*	Tipi di comunicazione con carta	*/
/*	CASE1: il reader non manda dati e non riceve dati*/
static int pt0_smartk_case1(APDU_Cmd *cmd, APDU_Rsp **rsp){
	int ret;
	BYTE buffer[5];
	APDU_Cmd *tpdu_cmd;
	
	PDEBUG("pt_t0:case1:ATTIVO\n");	
	
	/*	Riportiamoci nel caso di avere 0 byte da ricevere come da ISO 7816-3*/
	/*	APDU -> TPDU	*/
	memcpy (buffer, APDU_Cmd_Raw (cmd), 4);
  	buffer[4] = 0x00;

 	tpdu_cmd = APDU_Cmd_New (buffer, 5);

  	/* Invio TPDU */
  	ret = pt0_smartk_exchange_TPDU(tpdu_cmd, rsp);

  	/* Cancellazione TPDU */
  	APDU_Cmd_Delete (tpdu_cmd);

 	return ret;
}


/*	CASE2S: il reader non manda dati e riceve dati dalla carta.
	Ma invece questo sfaccimo ha fatto un errore! Ha switchare l'ISO
	quindi il reader manda i dati ma non riceve
 	Se Le=00 significa che si vuole ricevere il massimo dei dati(256)*/
static int pt0_smartk_case2S(APDU_Cmd *cmd, APDU_Rsp **rsp){
	
	PDEBUG("pt_t0:case2S:ATTIVO\n");
	
	/*	APDU->TPDU	(sono gli stessi)	*/
	/*	P3=Lc per indicare dati in arrivo	*/
	if(	pt0_smartk_exchange_TPDU(cmd, rsp) != OK){
		PERROR("pt_t0:case2S->Impossibile inviare comando!\n");
		return FAIL;
	}
	
	return OK;
}

/*	CASE3S: il reader non manda dati e riceve dati dalla carta	*/
static int pt0_smartk_case3S(APDU_Cmd *cmd, APDU_Rsp **rsp){
	int ret;
 	APDU_Rsp *tpdu_rsp;
 	BYTE buffer[5];
 	APDU_Cmd *tpdu_cmd;

	PDEBUG("pt_t0:case3S:ATTIVO\n");
	
	/*	Inviamo il comando di ricezione dati	*/
	ret = pt0_smartk_exchange_TPDU(cmd, &tpdu_rsp);
	
	/*	Gestione del comando	*/
	if (ret == OK){
		/*	Caso Le non accettato	*/
		if (APDU_Rsp_SW1 (tpdu_rsp) == 0x67){
			PERROR("pt_t0:case3S->Le non accettato\n");
      		(*rsp) = tpdu_rsp;
    	}

		/*	Case Le non accettato - ricevuto lunghezza differente dalla carta	*/
		else if (APDU_Rsp_SW1 (tpdu_rsp) == 0x6C){
			PDEBUG("pt_t0:case3S->SW1=0x6C - Richiesta Le=Na\n");
			/*	Nuovo comando per ricezione dati indicati da SW2	*/
			memcpy (buffer, APDU_Cmd_Raw (cmd), 4);
      		buffer[4] = APDU_Rsp_SW2 (tpdu_rsp);
	    	tpdu_cmd = APDU_Cmd_New (buffer, 5);
			
			/*	Invio del nuovo TPDU	*/
			APDU_Rsp_Delete(tpdu_rsp);
			ret = pt0_smartk_exchange_TPDU(tpdu_cmd, rsp);
			APDU_Cmd_Delete (tpdu_cmd);
			
			/*	Controllo della risposta	*/
			if( ret == OK){
				/*	Controllo dati superflui	*/
				if (APDU_Rsp_DataLen ((*rsp)) > APDU_Cmd_Le (cmd)){
        			APDU_Rsp_TruncateData ((*rsp), APDU_Cmd_Le (cmd));
					PERROR("pt_t0:case3S->SW1=0x6C - taglio dei dati superflui\n");
        		}
			}

		}
		/*	Inizio lettura dati con GET RESPONSE - fuori protocollo	*/
		else if(APDU_Rsp_SW1 (tpdu_rsp) == 0x61){
			PDEBUG("pt_t0:case3S->SW1=0x61 - Richiesta GET RESPONSE\n");
			/*	Preparazione della risposta	*/
			(*rsp) = tpdu_rsp;
			
			/*	Richiesta Get Response	*/
			buffer[0] = APDU_Cmd_Cla (cmd);
      		buffer[1] = 0xC0;
      		buffer[2] = 0x00;
      		buffer[3] = 0x00;
			
			/*Ricezione dei dati rimanenti	*/
			do{
	
				/*	Invio richiesta	*/
				buffer[4] = APDU_Rsp_SW2 (tpdu_rsp);
        		tpdu_cmd = APDU_Cmd_New (buffer, 5);
        		ret = pt0_smartk_exchange_TPDU(tpdu_cmd, &tpdu_rsp);
				APDU_Cmd_Delete (tpdu_cmd);
				
				/*	Concatenazione dei dati	*/
				if( ret == OK ){
					if (APDU_Rsp_AppendData ((*rsp), tpdu_rsp) != APDU_OK){
          				PERROR("pt_t0:case3S->SW1=0x61 - Impossibile concatenare dati della risposta\n");
						ret = FAIL;
					}
					
					/*Pulizia della memoria tanto viene allocata da exchange_TPDU	*/
					APDU_Rsp_Delete (tpdu_rsp);
				}
			}while ( (ret == OK) && (APDU_Rsp_SW1 (*rsp) == 0x61) );
			/*	Controllo sulla lunghezza della risposta	*/
			if( ret == OK){
				if (APDU_Rsp_DataLen ((*rsp)) > APDU_Cmd_Le (cmd))
        			APDU_Rsp_TruncateData ((*rsp), APDU_Cmd_Le (cmd));
				}
		}
		/*Le Ok - Aggiornamento della risposta	*/
		else{
    	PDEBUG("pt_t0:case3S->SW1=0x90 - Comando eseguito con successo\n");
			(*rsp) = tpdu_rsp;
		}
	}
	return ret;	
}


/*	CASE4S: il reader manda dati e riceve dati dalla carta	*/
static int pt0_smartk_case4S(APDU_Cmd *cmd, APDU_Rsp **rsp){
	/*	Riportiamoci nel caso2S */
	int ret;
	BYTE buffer[PROTOCOL_T0_MAX_SHORT_COMMAND];
	APDU_Cmd *tpdu_cmd;
  	APDU_Rsp *tpdu_rsp;

	PDEBUG("pt_t0:case4S:ATTIVO\n");
	
	/*	APDU -> TPDU */
	/*	Riportiamoci nel CASE 2	*/
  	memcpy (buffer, APDU_Cmd_Raw (cmd), APDU_Cmd_RawLen (cmd) - 1);
  	tpdu_cmd = APDU_Cmd_New (buffer, APDU_Cmd_RawLen (cmd) - 1);

  	/* Invio TPDU */
  	ret = pt0_smartk_exchange_TPDU(tpdu_cmd,&tpdu_rsp);
  
	/* Pulizia TPDU del comando	*/
  	APDU_Cmd_Delete (tpdu_cmd);

	/*	Gestione del get response	*/
	if(ret == OK){
		/*	Comando eseguito com dati aggiuntivi	*/
		if (APDU_Rsp_SW1 (tpdu_rsp) == 0x61){
			PDEBUG("pt_t0:case4S->SW1=0x61 - SW2=%02X) - Le=%02X\n", APDU_Rsp_SW2(tpdu_rsp), APDU_Cmd_Le(cmd));			
			
			/*	Preparazione comando get response	*/
			buffer[0] = APDU_Cmd_Cla (cmd);
      		buffer[1] = 0xC0;
      		buffer[2] = 0x00;
      		buffer[3] = 0x00;

		 	/*	Calcolo dei dati da richiedere	- Se 00 allora OK altrimenti MIN (Ne,Na)
				dove Ne sono i dati richiesti e Na quelli disponibili dalla risposta	*/
      		if (APDU_Rsp_SW2 (tpdu_rsp) == 0x00)
      			buffer[4] = (BYTE) APDU_Cmd_Le (cmd);
      		else
       			buffer[4] = MIN (APDU_Cmd_Le (cmd), APDU_Rsp_SW2 (tpdu_rsp));

			/*	Invio del nuovo comando	*/
      		tpdu_cmd = APDU_Cmd_New (buffer, 5);
			APDU_Rsp_Delete (tpdu_rsp);
			ret = pt0_smartk_exchange_TPDU(tpdu_cmd, rsp);
      		APDU_Cmd_Delete (tpdu_cmd);
		}
		/*	Comando non accettato	*/
		else if ((APDU_Rsp_SW1 (tpdu_rsp) & 0xF0) == 0x60){
			PERROR("pt_t0:case4S->SW1=0x6x - Comando Errato\n");			
			/*	prepariamo la sisposta	*/
			*rsp = tpdu_rsp;			
		}
		/*	Camando accettato e si aspetta la risposta	*/
		else{
			/*	Ricevuto 90 00 - Controllo di sicurezza	*/
			if ((APDU_Rsp_SW1 (tpdu_rsp) & 0xF0) != 0x90){
				PERROR("pt_t0:case4S->Ricevuto Errore Strano!\n");
				*rsp = tpdu_rsp;
				return FAIL;				
			}
			PDEBUG("pt_t0:case4S->SW1=90 - Invio get response per ricevere i dati\n");
			
			/*	Prepararazione Get Reponse TPDU	completo dopo la ricezione di 90 00*/
			APDU_Rsp_Delete (tpdu_rsp);
     		buffer[0] = APDU_Cmd_Cla (cmd);
     		buffer[1] = 0xC0;
     		buffer[2] = 0x00;
     		buffer[3] = 0x00;
     		buffer[4] = (BYTE) APDU_Cmd_Le (cmd);
     		tpdu_cmd = APDU_Cmd_New (buffer, 5);

	    	/*	Riportiamoci nel caso 3S poiche' dobbiamo ricevere dati dalla carta	*/
     		ret = pt0_smartk_case3S(tpdu_cmd, rsp);
			APDU_Cmd_Delete (tpdu_cmd);
		}

	}
	
	if(ret == OK)
		PDEBUG("pt_t0:case4S->len_resp = %d\n", APDU_Rsp_RawLen(*rsp));
	else
		PDEBUG("pt_t0:case4S->Errrrrrrrrorrrreeeeeeeeee!\n");
	return ret;
}

/*	Comunicazione con il lettore utilizzando TPDU*/
static int pt0_smartk_exchange_TPDU(APDU_Cmd *cmd, APDU_Rsp **rsp){
	BYTE buffer[PROTOCOL_T0_MAX_SHORT_RESPONSE];
  	BYTE *data, *inv_cmd;
 	long Lc, Le, sent, recv;
 	int ret = OK, nulls, cmd_case;

 	/* Info APDU */
  	Lc = APDU_Cmd_Lc (cmd);
  	Le = APDU_Cmd_Le (cmd);
  	data = APDU_Cmd_Data (cmd);
  	
	PDEBUG("pt_t0:exchange_TPDU->Lc = %02X - Le = 0x%02X %d\n", Lc, Le, Le);
	
	cmd_case = APDU_Cmd_Case (cmd);

  	/* Check case of command */
  	if ((cmd_case != APDU_CASE_2S) && (cmd_case != APDU_CASE_3S))
   		PERROR("pt_t0:exchange_TPDU-> Classe di comando errato!\n");
	
	/*	Controllo convention	*/
	if( pt0_smartk_check_begin_convention()!= OK){
		PERROR("pt_t0:exchange_TPDU->Impossibile controllare la convention\n");
		/*	Tentativo di riconfigurare la seriale	*/
		pt0_smartk_check_end_convention();
		return FAIL;		
	}
	
	PDEBUG("pt_t0:exchange_TPDU->AGGIORNATA CONVENTION\n");
		
	/*	Invio dati comunicazione	*/
	if( pt0_smartk_trasmit( APDU_Cmd_Header (cmd), 5) != OK ){
		PERROR("pt_t0:exchange_TPDU->Impossibile inviare il comando\n");
		return FAIL;
	}
	
	/*	Init contatori	*/
	nulls = 0;
  	sent = 0;
  	recv = 0;
	
	/*	Gestione della comucazione*/
	while (recv < PROTOCOL_T0_MAX_SHORT_RESPONSE){
		
		/*	Analisi primo byte di risposta del lettore	*/
		if( pt0_smartk_receive(buffer+recv, 1) != OK){
			PERROR("pt_t0:exchange_TPDU->Impossibile ricevere byte di controllo\n");
			ret = FAIL;
			break;			
		}
		
		/*	Caso NULL byte ovvero un modo per la carta per guadagnare 
		*	tempo sulla esecuzione del comando	*/
		if(buffer[recv] == 0x60){
			nulls ++;
			/*	Controllo max null byte	*/
			if( nulls > PROTOCOL_T0_MAX_NULLS){
				PERROR("pt_t0:exchange_TPDU->Raggiunta max numero di null:uscita\n");
				ret = FAIL;
				break;
			}
			continue;
		}
		/*	Caso SW1	*/
		else if( ((buffer[recv] & 0xF0) == 0x60) || ((buffer[recv] & 0xF0) == 0x90)){
			recv++;
			
			/*	Controllo max risposta	*/
			if (recv >= PROTOCOL_T0_MAX_SHORT_RESPONSE){
				PERROR("pt_t0:exchange_TPDU->Raggiunta Risposta massima:uscita\n");
				ret = FAIL;
				break;
			}
			
			/*	Lettura SW2	*/
         	if( pt0_smartk_receive(buffer+recv, 1) != 0){
				PERROR("pt_t0:exchange_TPDU->Impossibile ricevere byte di controllo\n");
				ret = FAIL;
				break;				
			}   	
			
			PDEBUG("pt_t0:Risposta SW2 = %02X\n", buffer[recv]);
			recv ++;
			
			/*	Lettura Risposta completata*/
			ret = OK;
			break; 
		}
		/*	Caso ACK*/
		else if( (buffer[recv] & 0x0E) == (APDU_Cmd_Ins (cmd) & 0x0E) ){
			
			PDEBUG("pt_t0:exchange_TPDU->Ricevuto ACK\n");
			/*	reset null	*/
			nulls = 0;
			
			/*	CASE2:reader -> card	*/
			if( cmd_case == APDU_CASE_2S ){
				/*	Controllo lunghezza trasmissione	*/
				if(sent >= Lc){
					PERROR("pt_t0:exchange_TPDU->Nbyte da mandare >= Lc!\n");
					ret = FAIL;
					break;
				}
				if( pt0_smartk_trasmit(data+sent, MAX(Lc - sent, 0)) != OK ){
					PERROR("pt_t0:exchange_TPDU->Impossibile trasmettere byte restanti!\n");
					ret = FAIL;
					break;
				}
			
				sent = Lc;
				continue;
			}
			/*	CASE3:card -> reader	*/
			else{
				PDEBUG("pt_t0:exchange_TPDU->Case 3 Ricezione dati MAX= %d!\n", MAX(Le - recv,0));
				/*	Controllo sulla ricezione	*/
				if(recv >= PROTOCOL_T0_MAX_SHORT_RESPONSE){
					PERROR("pt_t0:exchange_TPDU->Nbyte da ricevere illimitato!\n");
					ret = FAIL;
					break;
				}
				if( pt0_smartk_receive(buffer+recv, MAX(Le - recv,0)) != OK ){
					PERROR("pt_t0:exchange_TPDU->Impossibile ricevere byte restanti!\n");
					ret = FAIL;
					break;
				}
				recv = Le;
				continue;
			}
		}
		/*	ACK infelice = 1 byte alla volta	*/
		else if((buffer[recv] & 0x0E) == ((~APDU_Cmd_Ins (cmd)) & 0x0E)){
		/*	reset null	*/
			nulls = 0;
			
			/*	CASE2:reader -> card	*/
			if( cmd_case == APDU_CASE_2S ){
				/*	Controllo lunghezza trasmissione	*/
				if(sent >= Lc){
					PERROR("pt_t0:exchange_TPDU->Nbyte da mandare >= Lc!\n");
					ret = FAIL;
					break;
				}
				if( pt0_smartk_trasmit(data+sent, 1) != OK ){
					PERROR("pt_t0:exchange_TPDU->Impossibile trasmettere byte restanti!\n");
					ret = FAIL;
					break;
				}
				sent++;
				continue;
			}
			/*	CASE3:card -> reader	*/
			else{
				/*	Controllo sulla ricezione	*/
				if(recv >= PROTOCOL_T0_MAX_SHORT_RESPONSE){
					PERROR("pt_t0:exchange_TPDU->Nbyte da ricevere illimitato!\n");
					ret = FAIL;
					break;
				}
				if( pt0_smartk_receive(buffer+recv, 1) != OK ){
					PERROR("pt_t0:exchange_TPDU->Impossibile ricevere byte restanti!\n");
					ret = FAIL;
					break;
				}
				recv++;
				continue;
			}

		}
		/*	Caso INGESTIBILE	*/
		else{
			PERROR("pt_t0:exchange_TPDU->Ingestibile Valore! SW1=%02X\n", buffer[recv] );
			ret = OK;
			break;
		}
		
	}
	/*	Fine comunicazione	*/
	if( pt0_smartk_check_end_convention()!= OK){
		PERROR("pt_t0:exchange_TPDU->Impossibile riconfigurare la convention\n");
		return FAIL;		
	}
	/*	Preparazione Risposta e allocazione di memoria	*/
	if(ret == OK)
		*rsp = APDU_Rsp_New (buffer, recv);
	
	return ret;	
}


/*	Funzioni esterne del modulo	*/
static int pt0_smartk_init(void){
	
	/*	Controllo presenza modulo	*/
	/*	Da Fare	*/
	
	/*	Registrazione protocollo	*/
	if (register_pt_smartk(&t0) != OK){
		PERROR("pt_t0:init_card->Impossibile registrare il protocollo\n");
		return FAIL;
	}
	
	EXPORT_NO_SYMBOLS;
	PINFO("pt_t0:init->PT Layer On!\n");
	
	return OK;
}

static int pt0_smartk_init_card(struct smartk * core){
	PDEBUG("pt_t0:init_card!\n");
	/*	Aggiornamento core	*/
	mycore = core;
	
	/*	Allocazione Memoria	per ATR*/
	core->atr = ATR_New();
	char * raw_atr;
	int len_atr;
	
	/*	Attivazione carta	*/
	if (core->ifd == NULL ){
		PDEBUG("pt_t0:init_card:Cazzo non tengo il puntatore al lettore!\n");
		return FAIL;
	}
	PDEBUG("pt_t0:init_card-> ifd0=%X\n", (int)core->ifd);
	
	if (core->ifd->activate_card(core, TRUE, &raw_atr, &len_atr)){
		PERROR("pt_t0:init_card->Impossibile attivare la carta\n");
		return FAIL;		
	}
	
	/*	Raccolta informazioni utili della carta	*/
	/*	ATR	*/
	if( ATR_InitFromArray(core->atr, raw_atr, len_atr) != ATR_OK){
		PERROR("pt_t0:init_card->Impossibile analizzare ATR\n");
		return FAIL;
	}
	
	/*	Convenzione della carta	*/
	if( ATR_GetConvention(core->atr, &core->convention) != ATR_OK){
		PERROR("pt_t0:init_card->Impossibile calcolare convenzione\n");
		return FAIL;
	}
	
	PDEBUG("pt_t0:init_card->Detect %s convention card\n",\
		core->convention == ATR_CONVENTION_DIRECT ? "Direct" : "Inverse");
		
	/*	Setting Timings	- PPS	*/
	pt0_smart_set_timings();

	/*	Carta Attivata	*/
	core->is_present = TRUE;
	
	/*	Pulizia Memoria	per ATR	*/
	kfree(raw_atr);
	
	PINFO("pt_t0:init_card->SmartKard Ready!\n");
	return OK;
}


/*	Chiusura della carta	*/
static int pt0_smartk_cleanup_card(struct smartk * core){
	PDEBUG("pt_t0:cleanup_card!\n");
	/* Disattviamo la carta	*/
	if(core->ifd->deactivate_card(core) != OK){
		PERROR("pt_t0:cleanup_card->Impossibile disattivare la carta!\n");
		return FAIL;		
	}
	
	core->is_present = FALSE;
	
	/*	Liberiamo memoria occupata per ATR*/
	ATR_Delete(core->atr);
	mycore = NULL;
	
	PINFO("pt_t0:init_card->SmartKard Eject!\n");
	return OK;
}

/*	Invio e ricezione dei dati	*/
static int pt0_smartk_data(struct smartk * core, BYTE *new_cmd, int len_cmd, BYTE *new_rsp, int *len_rsp){
	/*	Creazione del comando da inviare	*/
	APDU_Cmd * cmd = APDU_Cmd_New(new_cmd, len_cmd);
	APDU_Rsp * rsp;
	
	int cmd_case, ret=OK;
	
	if( (cmd_case = APDU_Cmd_Case (cmd)) == APDU_MALFORMED ){
		PERROR("pt_t0:data->APDU comando errato!");
		return FAIL;		
	}
	
	PDEBUG("pt_t0:data->cmd=%s - len=%d CASE=%1d\n", mycore->xpdu_to_string(APDU_Cmd_Header(cmd), len_cmd), len_cmd, cmd_case);
	
	/*	Verifica tipo di comando da eseguire	*/
	switch(cmd_case){
		case APDU_CASE_1:
			ret = pt0_smartk_case1(cmd, &rsp); break;
		case APDU_CASE_2S:
			ret = pt0_smartk_case2S(cmd, &rsp); break;
		case APDU_CASE_3S:
			ret = pt0_smartk_case3S(cmd, &rsp); break;
		case APDU_CASE_4S:
			ret = pt0_smartk_case4S(cmd, &rsp); break;
		default:
			PERROR("pt_t0:data->Tipo di comando non riconosciuto! cmd_case = %d\n", cmd_case);
			ret = FAIL;
	}
	
	if(ret != OK){
		PERROR("pt_t0:data->Problemi nell'eseguire il comando!\n");
		return FAIL;
	}
	
	/*	Copia della risposta	*/
	(*len_rsp) = APDU_Rsp_RawLen(rsp);
	memcpy(new_rsp, APDU_Rsp_Data(rsp), APDU_Rsp_RawLen(rsp));	
	
	PDEBUG("pt_t0:data->Risultato finale: SW1=%02X SW2=%02X RSP_RawLen=%d\n", APDU_Rsp_SW1(rsp), APDU_Rsp_SW2(rsp), *len_rsp);
	
	return OK;
}


static void pt0_smartk_exit(void){
	//pt0_smartk_cleanup_card(mycore);
	printk(KERN_INFO "pt_t0:exit->Protocol T0 Layer OFF!\n");
}



module_init(pt0_smartk_init);
module_exit(pt0_smartk_exit);
