/*
	Implementation of SmartK architecture.
	
	This file is part of Smartk, a smart card framework for the Linux Kernel.
	Copyright (C) 2003	Luigi Catuogno		<luicat@dia.unisa.it>
						Roberto Gassira'	<robgas@dia.unisa.it>
						Michele Masullo		<micmas@dia.unisa.it>
	
    Smartk is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Smartk is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
	
#include "smartk.h"
#include <linux/mm.h>	/*x allocazione memoria (mettere come ultimo!)*/

/* Definizione macro per licenza */
#define MY_AUTHOR       "Roberto Gassira' <robgas@tcfs.it>"
#define MY_DESC         "Modulo di base per interfacciare la seriale"

/* Identificazione del codice e della sua licenza*/
MODULE_AUTHOR(MY_AUTHOR);
MODULE_DESCRIPTION(MY_DESC);
MODULE_LICENSE("GPL");

/*	Moduli di default da caricare	*/
//#define DEFAULT_PORT	serial;
//#define DEFAULT_READER	towitoko;

/*	Definizione oggetto globale	*/
static struct smartk core;

/*	Funzioni esportate dal modulo	*/
int register_io_smartk(struct io_smartk *new_io);
int register_ifd_smartk(struct ifd_smartk *new_ifd);
int register_pt_smartk(struct pt_smartk *new_pt);
int smartk_init_card();
int smartk_data(BYTE *new_cmd, int len_cmd, BYTE *new_rsp, int *len_rsp);
int smartk_cleanup_card();

/* Funzioni Interne*/
char * xpdu_to_string(BYTE *command, int len){
	int i, total = len*3+1;
	char * ret;
	/*	Allocazione memoria per stringa*/
	ret = (char *)kmalloc(total, GFP_KERNEL);
	for(i=0; i< len; i++)
		sprintf(ret+(i*3),"%02X ", command[i]);
	ret[total] = '\0';
	return ret;
}

/*	Funzioni Modulo	*/
int smartk_init(){
	/*	Init della struttura dati	*/
	core.atr = NULL;
	core.convention = IO_SERIAL_PARITY_NONE;
	core.is_present = FALSE;
	core.timings.block_delay = core.timings.char_delay = 0;
	core.xpdu_to_string = xpdu_to_string;
	core.io = NULL;
	core.ifd = NULL;
	core.pt = NULL;
	
	PINFO("Init... \n");
	return OK;
}

int register_io_smartk(struct io_smartk *new_io){
	/*	Aggiornamento dell'oggeto per porta comunicazione	*/
	core.io = new_io;
	
	PINFO("Register IO...\n");
	return OK;
}

int register_ifd_smartk(struct ifd_smartk *new_ifd){
	/*	Registrazione device driver lettore	*/
	core.ifd = new_ifd;
	
	PINFO("Register IFD...\n");
	PDEBUG("new_ifd=%X core->ifd=%X\n", (int)new_ifd, (int)core.ifd);
	
	/*	Attivazione del lettore	*/
	if( core.ifd->init_reader(&core) != OK){
		PERROR("Impossibile attivare il lettore!\n");
		return FAIL;		
	}
	
	return OK;
}

int register_pt_smartk(struct pt_smartk *new_pt){
	/*	Registrazione device driver lettore	*/
	core.pt = new_pt;
	
	PINFO("Register PT...\n");
	return OK;
}

/*	====> FrAmEwOrK <====	*/
int smartk_init_card(){
	PDEBUG("smartk_init_card!\n");
	/*	Controllo definito protocollo	*/
	if(core.pt == NULL){
		PERROR("Protocollo di comunicazione non definito!\n");
		return FAIL;
	}
	/*	Chiamata funzione protocollo	*/
	return( core.pt->activate_card(&core) );
}

int smartk_cleanup_card(){
	PDEBUG("disactivate_card!\n");
	/*	Controllo definito protocollo	*/
	if(core.pt == NULL){
		PERROR("Protocollo di comunicazione non definito!\n");
		return FAIL;
	}
	
	/*	Disattivazione Carta	*/
	return(core.pt->disactivate_card(&core));
}

int smartk_data(BYTE *new_cmd, int len_cmd, BYTE *new_rsp, int *len_rsp){
	/*	Controllo definito protocollo	*/
	if(core.pt == NULL){
		PERROR("Protocollo di comunicazione non definito!\n");
			return FAIL;
	}
	PDEBUG("Comando Arrivato = %s\n", core.xpdu_to_string(new_cmd, len_cmd));
	/*Invio comando*/
	return(core.pt->data(&core, new_cmd, len_cmd, new_rsp, len_rsp));
}


void smartk_cleanup(){
	/*	Disabilitiamo il lettore	*/
	PINFO("Exit...\n");

}

/*	Esportazione funzioni	*/
EXPORT_SYMBOL (register_io_smartk);
EXPORT_SYMBOL (register_ifd_smartk);
EXPORT_SYMBOL (register_pt_smartk);
EXPORT_SYMBOL (smartk_init_card);
EXPORT_SYMBOL (smartk_cleanup_card);
EXPORT_SYMBOL (smartk_data);

/*	Definizione init del modulo	*/
module_init(smartk_init);
module_exit(smartk_cleanup);
