/*
	Definition of basic structure.
	
	This file is part of Smartk, a smart card framework for the Linux Kernel.
	Copyright (C) 2003	Luigi Catuogno		<luicat@dia.unisa.it>
						Roberto Gassira'	<robgas@dia.unisa.it>
						Michele Masullo		<micmas@dia.unisa.it>
	
    Smartk is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Smartk is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/


#ifndef _SMARTK_H
#define _SMARTK_H

#undef PDEBUG		
#ifdef SMARTK_DEBUG
#	define PDEBUG(fmt, args...) printk(KERN_ALERT "smartk: " fmt, ##args)
#else
#	define PDEBUG(fmt, args...)	/*No debug messages*/ 
#endif /*end SMARTK_DEBUG*/

#undef PERROR
#define PERROR(fmt, args...) printk(KERN_ERR "smartk: " fmt, ##args)

#undef PINFO
#define PINFO(fmt, args...) printk(KERN_INFO "smartk: " fmt, ##args)

/* include comuni */
#include <linux/module.h>		/*x creazione del moduli		*/
#include <linux/kernel.h> 		/*x info kernel come printk		*/
#include <linux/init.h>			/*x funzioni module_init ecc.	*/
#include <asm/errno.h> 			/*x valori errori				*/
#include <linux/wrapper.h>		/*wrapper future versioni		*/
#include "base.h"
#include "atr.h"

//#if defined(CONFIG_MODVERSIONS)
#include <linux/modversions.h>	/*x la versione delle funzioni*/
//   #define MODVERSIONS
//#endif


/* macro comuni */
/* Configurazione della seriale */
#define IO_SERIAL_HIGH      	1
#define IO_SERIAL_LOW     		0
#define IO_SERIAL_PARITY_ODD            1
#define IO_SERIAL_PARITY_EVEN           2
#define IO_SERIAL_PARITY_NONE           3

#define MAXIODEVNAME			16
#define MAXREADERNAME			64
#define MAXMODNAME				16
#define VERSION 				1


/*	Gestione seriale	*/
struct io_property{
	unsigned long bitrate;				//Velocita' input e output
	unsigned bits;						//BitStop
	unsigned stopbits;
	int parity;
	int dtr;
	int rts;
};

/*	Timeout*/
struct timings_card{
	unsigned block_delay;          /* Delay (ms) after starting to transmit */
  	unsigned char_delay;           /* Delay (ms) after transmiting each sucesive char*/
  	unsigned block_timeout;        /* Max timeout (ms) to receive first char */
  	unsigned char_timeout;         /* Max timeout (ms) to receive sucesive characters */
};

/*	Oggetto Globale	*/
struct smartk{
	/*	Info smartcard	*/
	ATR * atr;
	int convention;
	short is_present;
	struct timings_card timings;
	
	/*	Puntatori agli oggetti della comunicazione	*/
	struct io_smartk * io;
	struct ifd_smartk * ifd;
	struct pt_smartk * pt;
		
	/*	Puntatori a funzioni di supporto	*/
	char* (*xpdu_to_string)(BYTE *command, int len);
};

/*	Struttura per IO	*/
struct io_smartk{
	char device[MAXIODEVNAME];
	struct io_property property;
		
	int (*init_port)(struct smartk * core);
	int (*cleanup_port)(struct smartk * core);
	int (*read)(struct smartk * core, BYTE * data, int count, unsigned timeout);
	int (*write)(struct smartk * core, BYTE * data, int count, unsigned timeout);
	int (*get_property)(struct smartk * core, struct io_property *newprop);
	int (*set_property)(struct smartk * core, const struct io_property *newprop);
};

/*	Struttura per lettore	*/
struct ifd_smartk{
	char reader[MAXREADERNAME];
	char port_mod[MAXMODNAME];
	int clock_rate;
	int parity;
	
	int (*init_reader)(struct smartk *core);
	int (*cleanup_reader)(struct smartk *core);
	int (*activate_card)(struct smartk *core, int with_ATR, char **res_ATR, int *len);
	int (*deactivate_card)(struct smartk *core);
	int (*set_parity)(struct smartk *core, int parity);
	int (*transmit)(struct smartk *core, BYTE* buffer, int count);
	int (*receive)(struct smartk *core, BYTE* buffer, int count);
};

/*	Struttura per il protocollo	*/
struct pt_smartk{
	int (*activate_card)(struct smartk * core);
	int (*data)(struct smartk * core, BYTE *new_cmd, int len_cmd, BYTE *new_rsp, int *len_rsp);
	int (*disactivate_card)(struct smartk * core);
};

#endif /*end _SMARTK_H*/
