/*
	Module test for user interface.
	
	This file is part of Smartk, a smart card framework for the Linux Kernel.
	Copyright (C) 2003	Luigi Catuogno		<luicat@dia.unisa.it>
						Roberto Gassira'	<robgas@dia.unisa.it>
						Michele Masullo		<micmas@dia.unisa.it>
	
    Smartk is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Smartk is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
	
#include <linux/modversions.h>

#include "base.h"
#include "smartk.h"
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/tty.h>
#include <asm/uaccess.h>  		/* for put_user */
#include <linux/mm.h>			

/* Definizione macro per licenza */
#define MY_AUTHOR       "bembo <robgas@tcfs.it>"
#define MY_DESC         "First version of my module"

/* Identificazione del codice e della sua licenza*/
MODULE_AUTHOR(MY_AUTHOR);
MODULE_DESCRIPTION(MY_DESC);
MODULE_LICENSE("GPL");

/*  Definizione Funzioni */
static int init_module(void);
static void cleanup_module(void);
static int device_open(struct inode *, struct file *);
static int device_release(struct inode *, struct file *);
static ssize_t device_read(struct file *, char *, size_t, loff_t *);
static ssize_t device_write(struct file *, const char *, size_t, loff_t *);
	
#define SUCCESS 0
#define DEVICE_NAME "smartk"/* Dev name as it appears in /proc/devices   */
#define DEVICE_MAJOR_NUMBER 241
#define BUF_LEN 1024           /* Max length of the message from the device */
	
#undef PDEBUG		
#define PDEBUG(fmt, args...) printk(KERN_ALERT "TEST_MOD: " fmt, ##args)
	
/* Global variables are declared as static, so are global within the file. */
	
static int Major;           /* Major number assigned to our device driver */
static int Device_Open = 0; /* Is device open?  Used to prevent multiple  */

/*	Buffer per la conservazione dei dati	*/
static struct {
	BYTE buffer[BUF_LEN];   
	int buff_len;
}data;
	
static struct file_operations fops = {
	.owner = THIS_MODULE,
	.read = device_read,
  	.write = device_write,
  	.open = device_open,
  	.release = device_release
};

/*	Funzioni interne	*/
/*	Funzione per trasformare una stringa in un BYTE */
BYTE * str_to_BYTE(char * data, int len_data, int *len_ret){
	BYTE * ret = (BYTE *) kmalloc((len_data/3+1) * sizeof(BYTE), GFP_KERNEL);
	int i,j=1, temp, value, k=0;
	
	for(i=0; i <= len_data; i++ ){
		//PDEBUG("hahah data[%d] = %c\n", i, data[i]);
		/*	Controllo spazio bianco	*/
		if((i % 3) == 2 ){
			//PDEBUG("AAGGIIOORRNNOO\n");
			j = 1;
			ret[k++] = value;
			value = 0;
			continue;
		}
		/*	Caso 0	*/
		if( (data[i] >= 48 && data[i] <= 57) )
			temp = data[i] - '0';
		/*	Caso A	*/
		else if (data[i] >= 65 && data[i] <= 70)
			temp = data[i] - 55;
		else continue;
			
		//value += (16 *data[i])<<(j-- * 4);
		value += temp * ((16*j--)?16:1);
		//PDEBUG("hahah data[%d] = %1X value = %d len_data = %d\n", i, temp, value, len_data);		
	}
	
	/*	Controllo sulla risposta	*/
	if(i != len_data+1){
		return NULL;	
	}
	
	*len_ret = k;
	
	//~ PDEBUG("len = %d\n", *len_ret);
	//~ for(i = 0; i < *len_ret; i++)
		//~ PDEBUG("ret[%d] = %02X\n", i,(int)ret[i]);
	
	
	return ret;
}


/* Funzioni per la gestione del modulo*/
/* Caricamento del modulo*/	
int init_module(void)
{
	//Registrazione del modulo nel kernel ovvero assegnazione di un major number	
	Major = register_chrdev(DEVICE_MAJOR_NUMBER, DEVICE_NAME, &fops);
	if (Major < 0) {
		printk (KERN_ALERT "Registering the character device failed with %d\n", DEVICE_MAJOR_NUMBER);
		return Major;
	}	
	
  	printk("TEST_MOD Loaded! Dev_Open = /dev/%s\n", DEVICE_NAME);

	return OK;
}


/* Scaricamento del modulo*/
void cleanup_module(void)
{	
	/* Unregister the device */
	int ret = unregister_chrdev(DEVICE_MAJOR_NUMBER, DEVICE_NAME);
	if (ret < 0) 
		printk("Error in unregister_chrdev: %d\n", ret);
	else
		PDEBUG("Unloaded!!\n");
}  
	
	
/* Funzioni per la gestione del character device  */
static int device_open(struct inode *inode, struct file *file)
{
	/*	Blocco per una solo lettura	*/
	if (Device_Open) 
		return -EBUSY;
	
	/*	Attivazione Carta */
	if(smartk_init_card() != OK){
		PDEBUG("Impossibile inizializzare la carta!\n");
		return FAIL;		
	}
	
	Device_Open++;
	
	printk(KERN_ALERT "TEST_MOD:Apertura device\n");
	MOD_INC_USE_COUNT;
	return SUCCESS;
}
	
	
/* Called when a process closes the device file.
 */
static int device_release(struct inode *inode, struct file *file)
{
	if(Device_Open)
		Device_Open --;     /* We're now ready for our next caller */

	MOD_DEC_USE_COUNT;
	
	/*	Disattivazione della carta	*/
	//smartk_cleanup_card();

	/*	Pulizia buffer	*/
	if(data.buff_len != 0){
		if( current->tty != NULL )
			current->tty->driver.write(current->tty, FALSE, data.buffer, data.buff_len);
		else
			printk(KERN_ALERT "TEST_MOD:Impossibile Output per utente %d\n", current->pid);
		
		data.buff_len = 0;
	}else
		smartk_cleanup_card();
	
	printk(KERN_ALERT "TEST_MOD:Chiusura device\n");
	return SUCCESS;
}
	
/*	Lettura del device	*/	
static ssize_t device_read(struct file *filp,
	char *buffer,    /* The buffer to fill with data */
	size_t length,   /* The length of the buffer     */
	loff_t *offset)  /* Our offset in the file       */
{
	/* Number of bytes actually written to the buffer */
	int bytes_read = 0;
	
	/* If we're at the end of the message, return 0 signifying end of file */
	if (data.buff_len == 0) 
		return 0;

	/*Log di connessione e studio dei vai vari cazzi della struct task*/
	PDEBUG("Collegato utente %d %s \n", current->uid, current->next_task->comm);

	/*Scrittura nella memoria utente*/
	if( copy_to_user(buffer, data.buffer,  data.buff_len) ){
		bytes_read = -EFAULT;
		PDEBUG("Impossibile scrivere nella memoria utente!");
		goto out;
	}
	
	/*	Aggiornamento valori letti	*/
	bytes_read = data.buff_len;
	data.buff_len = 0;
	
out:
   /* Most read functions return the number of bytes put into the buffer */
   return bytes_read;
}

	
/*  Called when a process writes to dev file: echo "hi" > /dev/hello */
static ssize_t device_write(struct file *filp,
	const char *buff,
	size_t len,
	loff_t *off)
{
	BYTE * cmd;
	BYTE resp[2];
	int ret=0, len_cmd, cazzo_resp;
	int * len_resp = (int *) kmalloc(sizeof(int), GFP_KERNEL);
	char temp[1024];
	
	/*Lettura dalla memoria utente*/
	if( copy_from_user(temp,buff, len)){
		ret = -EFAULT;
		goto out;
	}
	ret = len;
	cmd = str_to_BYTE(temp, len, &len_cmd);
	
	/*	Fine stringa comando	*/
	temp[len] = '\0';
	PDEBUG("From User dopo mapping :Data:%s Len:%d len_cmd=%d\n", temp, len, len_cmd);
	
	
	/*	Invio del comando	*/
	if ( smartk_data(cmd, len_cmd, resp, len_resp) != OK){
		PDEBUG("Errore comando\n");
		return FAIL;
	}

	PDEBUG("Risultato per len_resp di %d = %02X %02X\n", *len_resp, resp[(*len_resp)-2], resp[*(len_resp)-1]);	
	
	/*	Aggiornamento del buffer	*/
	memcpy(data.buffer, resp, *len_resp);
	data.buff_len = *len_resp;
	
	out:
	
	return ret;
}
